/**
 * Title: WinManager - A Java class managing creation of processes on the Windows platform.
 * Description:
 * Copyright:    Copyright (c) 2008
 * Company:      Morpheus
 * @author Stefano Adriani
 * @version 1.0
 */


package adriani.tutorial.system ;

import java.io.* ;
import java.lang.reflect.* ;


/**
* Class used to test the creation of new processes on the Windows platform. Following methods are provided:
*
* <H4>exec()</H4>
* This method does not provide a good management of the runtime exceptions, since the error message is not
* specified by the <CODE>Exception</CODE> object, but could depend by the operating system.
* For this reason, if the command file is not found, or the command path is wrong, the error message is evaluated
* assuming following error messages:<P>
*
* <UL>
* <LI><CODE>CreateProcess: <command> error=2</CODE>&nbsp;(file not found)</LI>
* <LI><CODE>CreateProcess: <path> error=3</CODE>&nbsp;(path not found)</LI>
* </UL>
*
* <STRONG>Input parameters:</STRONG>
* <UL>
* <LI><CODE>command</CODE>: specify the the command to be executed (as a <CODE>String</CODE> object).</LI>
* <LI><CODE>args[1]</CODE>: specify the mood of execution (0=NO_OUTPUT, 1=OUTPUT).</LI>
* <LI><CODE>args[2]</CODE>: specify how to handle the child (0=NO_WAIT, 1=WAIT, 2=KILL)</LI>
* </UL>
*
* <P>
*
* <H4>fork()</H4>
* In the Windows operating system is not easy to implement a system call like <CODE>fork()</CODE>,
* since usually this is available only as a C implicit library in the Linux enviroment.
* For this reason the <CODE>fork()</CODE> method offered by this class is a <I>fake</I> system call, since it
* does not realize a real <CODE>fork()</CODE> system call.<P>
*
*/
public class WinManager implements SystemManager {


	/** The absolute path of the installed Java Run Time (or JDK).													*/
	public String java_home ;


// *******************************			SETTING STATE			***********************************************


	/**
	* This method must be called after creation of the <CODE>SystemManager</CODE> instance, in order to
	* set the necessary Java Enviroment. For the time being following parameters must be specified:<P>
	*
	* <CODE>enviroment[0]</CODE>: the CLASSPATH variabile.
	*
	* @param enviroment an array specifying all the necessary enviroment's variables.
 	*/
	public void setState(String[] enviroment) {

	  if (enviroment != null) {
		this.java_home = enviroment[0] ;
	  }
	}


// *******************************			EMULATING EXE()		***********************************************


	/**
	* This method creates a new process invoking the <CODE>Runtime</CODE> object. The new process is executed as
	* a standalone application (not a new Thread).<BR>
	* Notice that the Windows version of this method is different from the Linux corresponding system call, since in
	* this case the segment code of the current program is NOT substituded!<P>
	*
	* <B>Remark:</B> the parameter param[0] is not used in current implementation (place holder).
	*
	* @param command the absolute path of the command (e.g. <CODE>C:\\WINDOWS\\system32\\cmd</CODE>).
	* @param param a <CODE>char</CODE> array specifying some parameters (see above).
	* @param verbose if <CODE>true</CODE> the process input is written on the command line (using decimal ASCII codes).
	* @return the string captured from the process's input.
 	*/
	public String exec(String command, char[] param, boolean verbose) {

	  String result = null ;
	  Process process = null ;
	  Runtime runtime = Runtime.getRuntime() ;

	  // 1. Creating the process
	  try {
	    process = runtime.exec(command) ;
	  }
	  catch(IOException e) {
	    int length = e.getMessage().length() ;
	    String end = e.getMessage().substring(length-1, length) ;
	    if (end.equals("2")) System.out.println("Errore: comando non trovato (" + command + ")") ;
	    else if (end.equals("3")) System.out.println("Errore: path non trovato (" + command + ")") ;
	    else System.out.println("Errore di I/O all'esecuzione: " + e.getMessage()) ;
	    System.exit(-1) ;
	  }
	  catch(Exception e) {
	    System.out.println("Errore generico all'esecuzione: " + e.getClass()) ;
	    System.exit(-1) ;
	  }

	  // 2. Showing results
	  System.out.println("Eseguito comando: " + command) ;

	  // NOTE: following code stops the current thread until the child gets his own input/output.
	  if (param[1] == '1') {
		result = "<input>" + getInput(process, verbose) + "</input>" ;
		System.out.println(result) ;
	  }

	  // NOTE: following code stops the current thread corrent until the child exits.
	  if (param[2] == '1') waitForChild(process) ;

	  // NOTE: if the child process has already terminated, this code returns an IllegalThreadStateException.
	  if (param[2] == '2') killChild(process) ;

	  return result ;
	}


// *******************************			EMULATING FORK()		***********************************************


	/**
	* This is an example of a wrong way to implement the <CODE>fork()</CODE> method on the Windows platform.
	* Here the input object (the <CODE>object</CODE> parameter) is used to create a new instance of the same type,
	* by calling the standard <CODE>object.newInstance()</CODE> method.<BR>
	* This implementation is wrong because in this way we just create a new instance of the input object, and not
	* a new copy of the parent process!<P>
	*
	* @param object the Java <CODE>object</CODE> represeting the parent process.
	* @param param a <CODE>char</CODE> array specifying some parameters (if necessary).
	* @param verbose if <CODE>true</CODE> some information about the parent object is printed on the console screen (using decimal ASCII codes).
	* @return the pid obtained after the clone() call (always -1, since this is not a real fork).
 	*/
	public int clone(Object obj, char[] param, boolean verbose) {

	  int pid = -1 ;
	  Object child = null ;
	  Class parent = obj.getClass() ;

	  // Applying introspection to the parent class
	  if (verbose) {
		  System.out.println("\nRisultato dell'introspezione su: " + parent.getName() + "\n") ;
		  System.out.println(getInformation(parent)) ;
		  System.out.println() ;
	  }

	  // Trying to clone the parent class
	  try {
		System.out.println("Eseguo clone di: " + obj) ;
		child = parent.newInstance() ;
		System.out.println("Creata classe: " + child) ;
	  } catch(Exception e) {
	    System.out.println("Errore di clonazione: " + e.getMessage()) ;
	  }

	  return pid ;
	}


	/**
	* This is an example of a fake implementation of the <CODE>fork()</CODE> method on the Windows platform.
	* This is not a real fork() call because here we create a copy of the parent process, but this is
	* not a good <CODE>fork()</CODE> call because the child is not an EXACT COPY of the parent, infact:
	*
	* <OL>
	* <LI>The PCB of the child is different from the PCB of the parent.</LI>
	* <LI>The Program Counter of the child process restarts from the begin.</LI>
	* <LI>After the fork() call, the child executes automatically an exec() call to the new process: this
	* means that we do not handle the result of the <CODE>fork()</CODE> call using the <CODE>(pid == 0)</CODE> mechanism.</LI>
	* </OL>
	*
	* <B>Remark:</B> the parameter param[0] is not used in current implementation (place holder).
	*
	* @param object the Java <CODE>object</CODE> represeting the parent process.
	* @param param a <CODE>char</CODE> array specifying some parameters (if necessary).
	* @param verbose if <CODE>true</CODE> the process input is printed on the console screen (using decimal ASCII codes).
	* @return the pid obtained after the fork() call (always -1, since this is not a real fork).
 	*/
	public int fork(Object obj, char[] param, boolean verbose) {

		int pid = -1 ;
		String command = null ;
		String argoments = null ;
		String child = obj.getClass().getName() ;
		if (obj instanceof adriani.tutorial.system.WinTester) argoments = " 0 3" ; else argoments = "" ;

		// Checking enviroment
		if (java_home == null) {
			System.out.println("WinManager: variabile java_home non settata") ;
			return -2 ;
		}
		// Executing a brand new process (not a real fork)
		command = java_home + " " + child + argoments ;
		String result = exec(command, param, verbose) ;

		return pid ;
	}


// *******************************			PROCESS MANAGEMENT		***********************************************


	/**
	* Wait for the child process.
	*
	* @param process the target process.
 	*/
	public void waitForChild(Process process) {

	  try {
		System.out.print("In attesa di: " + process + " ") ;
	    process.waitFor() ;
	    System.out.println("... (" + process.exitValue() + ")") ;
	  } catch(InterruptedException e) {
	    System.out.println("Errore in attesa: " + e.getMessage()) ;
	  }
	}


	/**
	* Kill the child process.
	*
	* @param process the target process.
 	*/
	public void killChild(Process process) {

	  try {
		System.out.println("Sopressione di: " + process) ;
	    process.destroy() ;
	  } catch(Exception e) {
	    System.out.println("Errore di soppressione: " + e.getMessage()) ;
	    System.out.println("Eccezione: " + e) ;
	  }
	}


// *******************************			UTILITIES		***********************************************


	/**
	* Reading the input of the process.<BR>
	* Notice that this method is NOT contempled by the <CODE>SystemManager</CODE> interface: for this reason,
	* if this class is casted back into a <CODE>SystemManager</CODE> object, it will not be possibile to invoke this method.
	* Anyway, in the Java paradigm this method will be still available for internal used (i.e. it works if invoked
	* by this class).
	*
	* @param process the target process.
	* @param echo if <CODE>true</CODE> the captured input is printed on the console (using decimal ASCII code).
	* @return the string captured from the process's input.
 	*/
	private String getInput(Process process, boolean echo) {

      int left ;
      int valueRead ;
      int counter = 0 ;
      String line = "" ;
	  InputStream is = process.getInputStream() ;
	  BufferedInputStream bis = new BufferedInputStream(is) ;

      try {
		if (echo) System.out.println("<byte>") ;
		while ( (valueRead = bis.read()) != -1 ) {
		  counter++ ;
		  left = bis.available() ;
		  line += (char) valueRead ;
		  if (echo) System.out.print(counter + ":" + valueRead + " ") ;
		  if (left == 0) break ;
		}
		if (echo) System.out.println("\n</byte>") ;

	  } catch(IOException e) {
	    System.out.println("Errore nella cattura dell'input: " + e.getMessage()) ;
	  } catch(Exception e) {
	    System.out.println("Errore generico al momento della cattura: " + e.getClass()) ;
	  }

	  return line ;
	}


	/**
	* This method retrieve some information about the input object, using the <CODE>java.reflect.*</CODE> methods.
	*
	* @param parent the <CODE>Class</CODE> of the object the analyzed.
	* @return the result of the introspection methods.
 	*/
	private String getInformation(Class parent) {

		String result = "" ;

		result += "\tCostruttori:\n" ;
		Constructor[] c_list =  parent.getConstructors() ;
		for (int i = 0; i < c_list.length; i++) {
			Class[] param = c_list[i].getParameterTypes() ;
			result += "\t" + c_list[i].getName() + " (" ;
			for (int j = 0; j < param.length; j++) result += param[j].getName() + " " ;
			result += ")\n" ;
		}
		result += "\n" ;

		result += "\tAttributi:\n" ;
		Field[] f_list =  parent.getFields() ;
		for (int m = 0; m < f_list.length; m++) {
			result += "\t" + f_list[m].getType() ;
			result += " " + f_list[m].getName() + "\n" ;
		}

		return result ;
	}


} // End of class